__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from subprocess import Popen, PIPE, STDOUT
import json


def read_json(filename):
    with open(filename, "r") as fd:
        data = json.loads(fd.read())
    return data


def run_command(cmd, sudo=False, stream=False):
    """run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    """
    stdout = PIPE if not stream else None
    if sudo is True:
        cmd = ["sudo"] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=stdout)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    t = output.communicate()[0], output.returncode
    output = {"message": t[0], "return_code": t[1]}

    if isinstance(output["message"], bytes):
        output["message"] = output["message"].decode("utf-8")

    return output



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from compspec.logger import logger
import os

here = os.path.dirname(os.path.abspath(__file__))


def get_facts(names):
    """
    Retrieve logic programs by name, if they exists.
    We do not continue if they do not!
    """
    if not isinstance(names, list):
        names = [names]

    logic_programs = []

    for name in names:
        # First attempt - a full path was provided
        if os.path.exists(name):
            logic_programs.append(name)
            continue

        # Second attempt - logic program provided here
        path = os.path.join(here, name)
        if not os.path.exists(path):
            logger.exit(f"{path} does not exist.")
        logic_programs.append(path)

    return logic_programs



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

__version__ = "0.0.1"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsoch@users.noreply.github.com"
NAME = "compspec"
PACKAGE_URL = "https://github.com/compspec/compspec"
KEYWORDS = "diff, comparison, composition, specification, abi, compatibility, symbols"
DESCRIPTION = "Specification and model for describing and comparing things."
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (("clingo", {"min_version": None}),)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)
INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import compspec.solver
from compspec.solver import fn
from .base import CompositionBase, FactGenerator


class Composition(CompositionBase):
    """
    A composition is simply facts about one graph (object of interest).
    It uses a simple FactGenerator under the hood, and does not add any
    extra logic program (unless the user requests it).
    """

    def __init__(self, g, out=None, namespace=None, quiet=False):
        self.driver = compspec.solver.PyclingoDriver(out=out)
        self.facts = SingleCorpusGenerator(g, namespace=namespace)
        self.set_verbosity(out, quiet)


class SingleCorpusGenerator(FactGenerator):
    """
    The SingleCorpusGenerator generates facts for one graph.
    """

    def __init__(self, g, namespace=None):
        self.g = g
        self.ns = namespace or "A"

    def setup(self, driver):
        """
        Setup data for one library
        """
        self.gen = driver
        self.gen.h1(f"Composition Namespace {self.ns}")

        # Set the library namespace
        self.gen.fact(fn.namespace(self.ns))
        self.generate_facts(self.g, self.ns)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import compspec.solver
from compspec.solver import fn
from .base import CompositionBase, FactGenerator


class Combination(CompositionBase):
    """
    A combination combines one or more graphs.
    """

    def __init__(self, out=None, quiet=False):
        self.driver = compspec.solver.PyclingoDriver(out=out)
        self.facts = CombinedFactsGenerator()
        self.set_verbosity(out, quiet)

    def add_graph(self, g, ns):
        """
        Add a graph to the namespace.
        """
        self.facts.add_graph(g, ns)


class CombinedFactsGenerator(FactGenerator):
    """
    The SingleCorpusGenerator generates facts for one graph.
    """

    def __init__(self):

        # Lookup of graphs by namespace
        self.graphs = {}

    def add_graph(self, g, ns):
        """
        Add a graph to the namespace
        """
        self.graphs[ns] = g

    def setup(self, driver):
        self.gen = driver
        for ns, g in self.graphs.items():
            self.gen.h1(f"Composition Namespace {ns}")
            self.gen.fact(fn.namespace(ns))
            self.generate_facts(g, ns)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import compspec.entity as entity


class Graph:
    """
    A graph implicitly is scoped to one namespace
    """

    def __init__(self):

        # A counter to keep track of ids in this space
        self.count = entity.get_counter()
        self.ids = {}
        self.nodes = {}
        self.lookup = {}
        self.relations = []

    def to_dict(self):
        """
        Output dictionary representation of nodes and relations.
        """
        return {
            "nodes": [v.to_dict() for _, v in self.nodes.items()],
            "relations": [x.to_dict() for x in self.relations],
        }

    @classmethod
    def from_dict(self, obj):
        """
        Return a new graph loaded from a dictionary.
        """
        nodes = obj.get("nodes", [])
        relations = obj.get("relations", [])
        g = Graph()
        [g.new_node(**x) for x in nodes]
        [g.new_relation(**x) for x in relations]
        return g

    def next(self):
        """
        Return next id in the counter.
        """
        return next(self.count)

    def add_node(self, node):
        """
        Add an already generated node.
        """
        self.nodes[f"{node.uid}"] = node

    def add_relation(self, relation):
        """
        Add an already generated relation.
        """
        self.relations.append(relation)

    def new_node(self, name, value, nodeid=None):
        """
        Generate a node with a name (type) and value
        """
        if not nodeid:
            nodeid = self.next()
        node = entity.node(nodeid, name, value)
        self.nodes[f"{nodeid}"] = node
        return node

    def new_relation(self, fromid, relation, toid):
        """
        Generate a relation between parent (fromid) and child (toid).
        The relation here does not hard code a namespace, but it will
        be required to compare between two graphs.
        """
        relation = entity.relation(fromid, relation, toid)
        self.relations.append(relation)
        return relation

    def gen(self, name, value, parent, nodeid=None, relation="has"):
        """
        Generate a node and relation in one swoop!
        A parent is required.
        """
        if not nodeid:
            nodeid = self.next()
        node = entity.node(nodeid, name, value)
        self.nodes[nodeid] = node
        relation = self.new_relation(parent, relation, node.nodeid)
        self.relations.append(relation)
        return node, relation

    def iter_nodes(self):
        """
        Yield nodes. If a comparison is being done, a namespace needs to be
        added (e.g., node, namespace, *args)
        """
        for _, node in self.nodes.items():
            yield node.args

    def iter_relations(self):
        """
        Yield relations in the same manner.
        """
        for relation in self.relations:
            yield relation.args



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


class Counter:
    """
    A simple counter (iterator) that will yield the next number with some prefix.
    """

    def __init__(self, prefix="id"):
        self.prefix = prefix

    def __next__(self):
        self.count += 1
        return "%s%s" % (self.prefix, self.count)

    def __iter__(self):
        # Actually start counting at 0
        self.count = -1
        return self


def get_counter(prefix="id"):
    """
    Return an iterable counter.
    """
    counter = Counter(prefix)
    return iter(counter)


class relation:
    """
    A relation in a graph.
    """

    def __init__(self, fromid, toid, relation="has"):
        self.fromid = fromid
        self.toid = toid
        self.relation = relation

    def __iter__(self):
        return iter((self.fromid, self.relation, self.toid))

    def to_dict(self):
        return {"fromid": self.fromid, "relation": self.relation, "toid": self.toid}

    def __str__(self):
        return f"{self.fromid}.{self.relation}:{self.toid}"

    def __repr__(self):
        return str(self)

    @property
    def args(self):
        return self.fromid, self.relation, self.toid


class node:
    """
    A node in a graph.
    """

    def __init__(self, nodeid, name, value):
        self.nodeid = nodeid
        self.name = name
        self.value = value

    def __str__(self):
        return f"{self.nodeid}.{self.name}:{self.value}"

    def __iter__(self):
        return iter((self.nodeid, self.name, self.value))

    def __repr__(self):
        return str(self)

    def to_dict(self):
        return {"nodeid": self.nodeid, "name": self.name, "value": self.value}

    @property
    def args(self):
        return self.nodeid, self.name, self.value



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import compspec.solver
from compspec.solver import fn
from .base import CompositionBase, FactGenerator


class Difference(CompositionBase):
    """
    A composition is simply facts about one graph (object of interest).
    It uses a simple FactGenerator under the hood, and does not add any
    extra logic program (unless the user requests it).
    """

    _logic_programs = ["is-compatible.lp"]

    def __init__(self, A, B, out=None, namespaceA=None, namespaceB=None, quiet=False):
        self.driver = compspec.solver.PyclingoDriver(out=out)
        self.facts = DiffFactsGenerator(
            A, B, namespaceA=namespaceA, namespaceB=namespaceB
        )
        self.set_verbosity(out, quiet)


class DiffFactsGenerator(FactGenerator):
    """
    The DiffFactsGenerator generates facts for two graphs to compare.
    """

    def __init__(self, A, B, namespaceA=None, namespaceB=None):
        self.A = A
        self.B = B
        self.nsA = namespaceA or "A"
        self.nsB = namespaceB or "B"

    def setup(self, driver):
        """
        Setup data for one library
        """
        self.gen = driver
        self.gen.h1(f"Difference Betweeen {self.nsA} and {self.nsB}")

        # Set the library namespace
        self.gen.fact(fn.is_a(self.nsA))
        self.gen.fact(fn.is_b(self.nsB))
        self.gen.h2(f"Namespace {self.nsA}")
        self.generate_facts(self.A, self.nsA)
        self.generate_facts(self.B, self.nsB)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import compspec.logger as logger
import clingo
import os

clingo_cffi = hasattr(clingo.Symbol, "_rep")


def _id(thing):
    """
    Quote string if needed for it to be a valid identifier.
    """
    if isinstance(thing, AspFunction):
        return thing
    elif isinstance(thing, int):
        return str(thing)
    # boolean and other casese
    return '"%s"' % str(thing)


def argify(arg):
    """
    Convert an argument into a clingo one.
    """
    if isinstance(arg, bool):
        return clingo.String(str(arg))
    elif isinstance(arg, int):
        return clingo.Number(arg)
    return clingo.String(str(arg))


class AspFunction:
    """
    An asp function
    """

    def __init__(self, name, args=None):
        self.name = name
        self.args = [] if args is None else args

    def __call__(self, *args):
        return AspFunction(self.name, args)

    def symbol(self, positive=True):
        return clingo.Function(
            self.name, [argify(arg) for arg in self.args], positive=positive
        )

    def __getitem___(self, *args):
        self.args[:] = args
        return self

    def __str__(self):
        return "%s(%s)" % (self.name, ", ".join(str(_id(arg)) for arg in self.args))

    def __repr__(self):
        return str(self)


class AspFunctionBuilder(object):
    def __getattr__(self, name):
        return AspFunction(name)


fn = AspFunctionBuilder()


class Result:
    """
    Result of an ASP solve.
    """

    def __init__(self, asp=None):
        self.asp = asp
        self.satisfiable = None
        self.optimal = None
        self.warnings = None
        self.nmodels = 0

        # specs ordered by optimization level
        self.answers = []
        self.cores = []


class PyclingoDriver:
    def __init__(self, cores=True, out=None):
        """
        Driver for the Python clingo interface.
        Arguments:
            cores (bool): whether to generate unsatisfiable cores for better
                error reporting.
            out (file-like): optional stream to write a text-based ASP program
                for debugging or verification.
        """
        self.set_verbosity(out)
        self.cores = cores

    def set_verbosity(self, out):
        """
        Set (or update) verbosity or output stream.
        """
        if out:
            self.out = out
        else:
            self.devnull()

    def devnull(self):
        self.f = open(os.devnull, "w")
        self.out = self.f

    def __exit__(self):
        self.f.close()

    def title(self, name, char):
        self.out.write("\n")
        self.out.write("%" + (char * 76))
        self.out.write("\n")
        self.out.write("%% %s\n" % name)
        self.out.write("%" + (char * 76))
        self.out.write("\n")

    def h1(self, name):
        self.title(name, "=")

    def h2(self, name):
        self.title(name, "-")

    def comment(self, msg):
        self.out.write("% " + msg + "\n")

    def newline(self):
        self.out.write("\n")

    def fact(self, head):
        """
        ASP fact (a rule without a body).
        """
        symbol = head.symbol() if hasattr(head, "symbol") else head
        self.out.write("%s.\n" % str(symbol))
        atom = self.backend.add_atom(symbol)
        self.backend.add_rule([atom], [], choice=self.cores)
        if self.cores:
            self.assumptions.append(atom)

    def solve(
        self,
        setup,
        nmodels=0,
        stats=False,
        logic_programs=None,
        facts_only=False,
    ):
        """
        Run the solver for a model and some number of logic programs
        """
        # logic programs to give to the solver
        logic_programs = logic_programs or []
        if not isinstance(logic_programs, list):
            logic_programs = [logic_programs]

        # Initialize the control object for the solver
        self.control = clingo.Control()
        self.control.configuration.solve.models = nmodels
        self.control.configuration.asp.trans_ext = "all"
        self.control.configuration.asp.eq = "5"
        self.control.configuration.configuration = "tweety"
        self.control.configuration.solve.parallel_mode = "2"
        self.control.configuration.solver.opt_strategy = "usc,one"

        # set up the problem -- this generates facts and rules
        self.assumptions = []
        with self.control.backend() as backend:
            self.backend = backend
            setup.setup(self)

        # If we only want to generate facts, cut out early
        if facts_only:
            return

        # read in provided logic programs
        for logic_program in logic_programs:
            self.control.load(logic_program)

        # Grounding is the first step in the solve -- it turns our facts
        # and first-order logic rules into propositional logic.
        self.control.ground([("base", [])])

        # With a grounded program, we can run the solve.
        result = Result()
        models = []  # stable models if things go well
        cores = []  # unsatisfiable cores if they do not

        def on_model(model):
            models.append((model.cost, model.symbols(shown=True, terms=True)))

        # Won't work after this, need to write files
        solve_kwargs = {
            "assumptions": self.assumptions,
            "on_model": on_model,
            "on_core": cores.append,
        }
        if clingo_cffi:
            solve_kwargs["on_unsat"] = cores.append
        solve_result = self.control.solve(**solve_kwargs)

        # once done, construct the solve result
        result.satisfiable = solve_result.satisfiable

        def stringify(x):
            if clingo_cffi:
                # Clingo w/ CFFI will throw an exception on failure
                try:
                    return x.string
                except RuntimeError:
                    return str(x)
            else:
                return x.string or str(x)

        if result.satisfiable:
            min_cost, best_model = min(models)
            result.answers = {}
            for sym in best_model:
                if sym.name not in result.answers:
                    result.answers[sym.name] = []
                result.answers[sym.name].append([stringify(a) for a in sym.arguments])

        elif cores:
            symbols = dict((a.literal, a.symbol) for a in self.control.symbolic_atoms)
            for core in cores:
                core_symbols = []
                for atom in core:
                    sym = symbols[atom]
                    core_symbols.append(sym)
                result.cores.append(core_symbols)

        if stats:
            print("Statistics:")
            logger.info(self.control.statistics)
        return result



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from compspec.solver import fn
import compspec.lp
import sys


class CompositionBase:
    """
    A composition base is the base for a Composition or Diff.
    """

    def set_verbosity(self, out, quiet=False):
        """
        Set the verbosity
        """
        out = out if out else sys.stdout
        if quiet:
            out = None
        self.out = out
        self.driver.set_verbosity(out)

    def solve(self, logic_programs=None):
        """
        Run the solve, optionally with extra logic programs.
        """
        result = self.driver.solve(self.facts, logic_programs=logic_programs)
        return result.answers

    def _load_logic_programs(self, logic_programs=None):
        """
        Load user- and class- provided logic programs.
        """
        # This will ensure any provided logic programs also exist
        logic_programs = logic_programs or []
        if isinstance(logic_programs, str):
            logic_programs = [logic_programs]
        logic_programs += getattr(self, "_logic_programs", [])
        if logic_programs:
            logic_programs = compspec.lp.get_facts(logic_programs)
        return logic_programs

    def run(self, logic_programs=None, quiet=False):
        """
        Run of a composition will output ASP facts, unless a logic program
        is provided then we do this full solve.
        """
        logic_programs = self._load_logic_programs(logic_programs)
        return self.solve(logic_programs=logic_programs)


class FactGenerator:
    """
    The FactGenerator generates facts for one graph.
    """

    def generate_facts(self, g, ns):
        """
        Generate facts for a namespaced graph
        """
        for relation in g.iter_relations():
            self.gen.fact(fn.relation(ns, *relation))
        for node in g.iter_nodes():
            self.gen.fact(fn.node(ns, *node))
